#!/usr/bin/env python
# Copyright 2015 Cisco Systems, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
#
# ---------------------------------------------------------
import bokeh.plotting
import os
import webbrowser

# Standard title attribute for all figures
title_style = {'title_text_font_size': '12pt',
               'title_text_font_style': 'bold'}

grid_title_style = {'title_text_font_size': '10pt',
                    'title_text_font_style': 'bold'}

def set_html_file(cdict_file, headless, label, output_dir):
    '''Sets the final html file name prefix and output directory
    if output_dir is None then use same output directory as cdict_file (can be relative)
    else use that directory

    prefix is set as following:
    if label is None use cdict basename, remove the .cdict extension and append the task_re
    else use label as prefix and do not append the task_re

    :param cdict_file: e.g. ../../perf.cdict
    :param headless:
    :param label: will replace all space with _
    :param output_dir:
    :return:
    '''
    global output_chart
    global headless_mode

    headless_mode = headless

    # the full prefix of the output file with directory pathname
    global output_file_prefix
    global ignore_task_re

    if headless:
        output_chart = bokeh.plotting.save
    else:
        output_chart = bokeh.plotting.show

    if label:
        ignore_task_re = True
        output_file_base = label.replace(' ', '-')
    else:
        ignore_task_re = False
        output_file_base = os.path.basename(cdict_file).replace('.cdict', '')

    if output_dir:
        output_file_dir = output_dir
    else:
        output_file_dir = os.path.dirname(cdict_file)

    if output_file_dir and output_file_dir[-1] != '/':
        output_file_dir += '/'
    output_file_prefix = output_file_dir + output_file_base

# calculate the time between the 1st entry and the last entry in msec
def get_time_span_usec(df):
    min_usec = df['usecs'].iloc[0]
    max_usec = df['usecs'].iloc[-1]
    return max_usec - min_usec

# For sorting
# 'CSR.1.vcpu0' => '0001.CSR.vcpu0'
def normalize_task_name(task):
    tokens = task.split('.')
    if len(tokens) > 1:
        # extract the service chain
        chain = tokens[1]
        try:
            chain = int(tokens[1])
            res = '%02d' % (chain)
            tokens.pop(1)
            return res + '.' + '.'.join(tokens)
        except ValueError:
            pass
    return task

def normalize_df_task_name(df):
    df['task_name'] = df.apply(lambda row: normalize_task_name(row['task_name']), axis=1)

def get_disc_size(count):
    if count < 100:
        return 10
    if count < 500:
        return 8
    return 6

def get_output_file_name(chart_type, task_re):
    filename = output_file_prefix + '-' + chart_type
    if ignore_task_re:
        filename += '.html'
    else:
        if task_re.startswith('.*'):
            # skip leading .* if present
            task_re = task_re[2:]
        filename += '_' + task_re + '.html'
    return filename

def output_html(chart, chart_type, task_re):
    filename = get_output_file_name(chart_type, task_re)
    bokeh.plotting.output_file(filename)
    print('Saved to ' + filename)
    output_chart(chart)

def output_svg_html(svg_html, chart_type, task_re):
    filename = get_output_file_name(chart_type, task_re)
    with open(filename, 'w') as dest:
        dest.write(svg_html)
        print('Saved to %s (%d Kbytes)' % (filename, len(svg_html) / 1000))
        if not headless_mode:
            # bring up the file in the default browser
            url = 'file://' + os.path.abspath(filename)
            webbrowser.open(url, new=2)

def get_full_task_name(df, task):
    # if task is a number it is considered to be a pid ID
    # if text it is a task name
    try:
        tid = int(task)
        # tid given
        df = df[df['pid'] == tid]
        task = None
        # get the task name from the tid
    except ValueError:
        # task given: find corresponding tid
        df = df[df['task_name'] == task]
        tid = 0

    if df.empty:
        print 'No selection matching the task ' + task
        return (None, None)
    # fill in the missing information
    if not tid:
        tid = df['pid'].iloc[0]
    if not task:
        task = df['task_name'].iloc[0]
    task = task + ':' + str(tid)
    return (df, task)

class DfDesc(object):
    '''A class to store a dataframe and its metadata:
    - time constrained df
    - multiplier (to indicate if the df is under sampled)
    - name
    '''
    def __init__(self, cdict_file, df):
        # remove the cdict extension if any
        if cdict_file.endswith('.cdict'):
            cdict_file = cdict_file[:-6]
        self.name = cdict_file
        self.multiplier = 1.0
        self.df = df
        self.short_name = cdict_file

    def normalize(self, from_time_usec, cap_time_usec):
        # remove all samples that are under the start time
        if from_time_usec:
            self.df = self.df[self.df['usecs'] >= from_time_usec]
        last_time_usec = self.df['usecs'].iloc[-1]
        # usec worth of samples in the df
        df_span_usec = last_time_usec - from_time_usec
        if cap_time_usec > df_span_usec:
            # eg if the requested cap is 1 sec and the df only contains
            # 500 msec of samples, the multiplier is 2.0
            self.multiplier = float(cap_time_usec) / df_span_usec
        # remove all samples that are over the cap
        self.df = self.df[self.df['usecs'] <= (cap_time_usec + from_time_usec)]



#!/usr/bin/env python
# Copyright 2015 Cisco Systems, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
#
# Author: Alec Hothan
# ---------------------------------------------------------

import pandas

from perfmap_core import get_cpu_sw_map

import itertools
import matplotlib.pyplot as plt
from matplotlib import colors
import numpy as np

# Use the matplotlib pastel1 palette for exit codes that do not have an assigned color
default_palette = [colors.rgb2hex(rgb) for rgb in plt.cm.Pastel1(np.linspace(0, 1, 10))]
default_color_palette = itertools.cycle(default_palette)
# This palette is extracted from colorbrewer2 qualitative palette with 10 colors
assigned_palette = ['#a6cee3', '#1f78b4', '#ff7f00', '#33a02c', '#fb9a99',
                    '#e31a1c', '#fdbf6f', '#b2df8a', '#cab2d6', '#6a3d9a']
assigned_index = 0
def assigned_color():
    global assigned_index
    res = assigned_palette[assigned_index]
    # limited to 10 for now
    assigned_index += 1
    return res

# KVM exit reasons
# Intel64 and IA32 Architecture Software Developer's Manual Vol 3B, System Programming Guide Part 2
# Appendix I
# The key is the numeric exit reason value
# The value is a list containing the exit reason clear text and an assigned color
# Most common exits are assigned a fixed color in order to avoid uneven color assignment across charts
# Exits without any assigned color will get a random assignment from a fixed color palette
# (with some possibly of near match with assigned colors unfortunately)
KVM_EXIT_REASONS = {
    0: ['Exception or NMI', assigned_color()],    #
    1: ['External Interrupt', assigned_color()],  #
    2: ['Triple Fault'],
    3: ['INIT'],
    4: ['Startup IPI'],
    5: ['I/O SMI (System Mgmt Interrupt)'],
    6: ['Other SMI'],
    7: ['Interrupt Window'],
    8: ['NMI window'],
    9: ['Task Switch'],
    10: ['CPUID'],
    11: ['GETSEC'],
    12: ['HLT', assigned_color()],     #
    13: ['INVD'],
    14: ['INVLPG'],
    15: ['RDPMC'],
    16: ['RDTSC'],
    17: ['RSM'],
    18: ['VMCALL'],
    19: ['VMCLEAR'],
    20: ['VMLAUNCH'],
    21: ['VMPTRLD'],
    22: ['VMPTRST'],
    23: ['VMREAD'],
    24: ['VMRESUME'],
    25: ['VMWRITE'],
    26: ['VMXOFF'],
    27: ['VMXON'],
    28: ['CR Access', assigned_color()],        #
    29: ['MOV DR'],
    30: ['I/O Instruction'],
    31: ['RDMSR'],
    32: ['WRMSR', assigned_color()],            #
    33: ['VM Entry Failure (invalid guest state)'],
    34: ['VM Entry Failure (MSR loading)'],
    35: ['n/a 35'],
    36: ['MWAIT'],
    37: ['Monitor trap flag'],
    38: ['n/a 38'],
    39: ['MONITOR'],
    40: ['PAUSE'],
    41: ['VM Entry Failure (machine check)'],
    42: ['n/a 42'],
    43: ['TPR below threshold'],
    44: ['APIC Access', assigned_color()],       #
    45: ['n/a 45'],
    46: ['Access to GDTR or IDTR'],
    47: ['Access to LDTR or TR'],
    48: ['EPT violation', assigned_color()],     #
    49: ['EPT misconfiguration'],
    50: ['INVEPT'],
    51: ['RDTSCP'],
    52: ['VMX preemption timer expired'],
    53: ['INVVPID'],
    54: ['WBINVD'],
    55: ['XSETBV'],
    56: ['APIC_WRITE', assigned_color()]         #
}


def aggregate_dfs(dfds, task_re):
    '''
    Aggregate dfs from multiple cdicts into 1 df that has the task names
    sufficed with the cdict name
    :param dfds: a list of df desc
    :param task_re: regex for task names
    :return: a tuple made of
        an aggregated df with annotated task names
        a dict of multipliers indexed by the annotated task name, corresponding to the
               ratio between the requested cap time and the cdict cap time (always >= 1.0)
               that should be used to adjust counts
    '''
    adjust_count_ratios = {}
    # annotate the task name with the cdict ID it comes from
    dfl = []
    for dfd in dfds:
        df = dfd.df
        df = df[df['event'] == 'kvm_exit']
        df = df[df['task_name'].str.match(task_re)]
        # add the cdict name to the task name unless there is only 1 cdict file
        if len(dfds) > 1:
            df['task_name'] = df['task_name'].astype(str) + '.' + dfd.short_name
        # check the time span
        if dfd.multiplier >= 1.01:
            print
            print 'Warning: counts for %s will be multiplied by %f' % \
                  (dfd.name, dfd.multiplier)
            # all these task names require a count adjustment
            adjust_task_names = pandas.unique(df.task_name.ravel()).tolist()
            for atn in adjust_task_names:
                adjust_count_ratios[atn] = dfd.multiplier
        dfl.append(df)
    df = pandas.concat(dfl)
    return df, adjust_count_ratios

def get_exit_color(code):
    exit_desc = KVM_EXIT_REASONS[code]
    try:
        color = exit_desc[1]
    except IndexError:
        # unassigned color, assign one at runtime from a default color palette map
        color = str(next(default_color_palette))
        exit_desc.append(color)
    return color

'''
def convert_exit_df(df, label):
    # fill in the error reason text from the code in
    df.next_comm = df.next_comm.apply(lambda x: '(%02d) %s' % (x, KVM_EXIT_REASONS[x][0]))
    series_percent = df.next_comm.value_counts(normalize=True)
    series_count = df.next_comm.value_counts()
    series_percent = pandas.Series(["{0:.2f}%".format(val * 100) for val in series_percent],
                                   index=series_percent.index)

    series_percent.name = '%'
    series_count.name = label + ' count'

    res = pandas.concat([series_count, series_percent], axis=1)
    return res

    # Show aggregated time inside VM and inside KVM
    print
    print 'Aggregated duration:'
    for event in legend_map:
        percent = accumulated_time[event] * 100 / total_time
        print '   %40s: %9d usec %3d%%' % (legend_map[event][1], accumulated_time[event], percent)
    print '   %40s: %9d usec %3d%%' % ('Total', total_time, 100)
    print

    # all exit types
    dfexits = df[df.event == 'kvm_exit']

    # now keep only the last exit before a switch
    # only keep kvm_exits and context switches
    df = df[(df['event'] == 'kvm_exit') | (df['event'] == 'sched__sched_switch')]

    # move all rows up
    sdf = df.shift(-1)
    # only keep rows that have a different event than their next row
    df = df[df['event'] != sdf['event']]
    # delete last row
    df = df.drop(df.tail(1).index)

    # delete all sched switches as they have served their purpose
    df = df[df['event'] == 'kvm_exit']

    show_exit_type_count(dfexits, df)

    def show_exit_type_count(df_all_exits, df_last_exits):

        res_all = convert_exit_df(df_all_exits, 'total')
        res_last = convert_exit_df(df_last_exits, 'last exit')
        res = pandas.concat([res_all, res_last], axis=1)
        res.fillna(0, inplace=True)
        res.sort_values('total count', inplace=True, ascending=False)
        print res
'''


def get_kvm_exit_data(dfds, cap_time_usec, task_re):

    # calculate the total cpu and total context switches per task
    cpu_sw_map = get_cpu_sw_map(dfds, cap_time_usec, task_re)

    # a dict of adjustment ratios indexed by task name for cdicts
    # that require count adjustment due to
    # capture window being too small
    df, adjust_count_ratios = aggregate_dfs(dfds, task_re)
    if df.empty:
        print 'Error: No kvm traces matching ' + task_re
        return

    df.drop(['cpu', 'duration', 'event', 'next_pid', 'pid', 'usecs'], inplace=True, axis=1)

    # Get the list of exit reason codes, sorted numerically
    exit_code_list = pandas.unique(df.next_comm.ravel()).tolist()
    exit_code_list.sort()

    # key = exit code, value = exit index
    exit_index_map = {}
    # index is the exit index
    exit_reason_list = []
    #     colormap_list = [
    #    {"exit":"EPT violation", "code":"#b2df8a"},
    #    {"exit":"VMRESUME", "code":"#b2df8a"},
    #    {"exit":"HLT", "code":"#b2df8a"}]
    colormap_list = []
    exit_index = 0
    for exit_code in exit_code_list:
        try:
            exit_reason_list.append(KVM_EXIT_REASONS[exit_code][0])
            colormap_list.append(get_exit_color(exit_code))
            exit_index_map[exit_code] = exit_index
            exit_index += 1
        except KeyError:
            # one of the next_comm row has an un-mapped exit reason code
            # bacause our map key range is contiguous, it must be a value out of range
            print 'Error: KVM exit reason %d unknown' % (exit_code)
            return

    # group by task name then exit reasons
    gb = df.groupby(['task_name', 'next_comm'])
    # number of exit types in each group
    # result is a series with 2-level index (task_name, next_comm)
    size_series = gb.size()

    # Get the list of all level 0 indices
    task_names = size_series.index.levels[0]

    #     task_list = [
    #    {"name":"Router", "exit_list":[{"name":"EPT violation", "count":400}, {"name":"APIC_WRITE", "count":300}]},
    #    {"name":"Firewall", "exit_list":[{"name":"VMRESUME", "count":300}, {"name":"HLT", "count":930}]}]
    task_list = []
    for task_name in task_names:
        exit_count_list = [0] * len(exit_reason_list)
        if adjust_count_ratios:
            try:
                multiplier = adjust_count_ratios[task_name]
            except KeyError:
                multiplier = 1
        else:
            multiplier = 1
        for exit_code, count in size_series[task_name].iteritems():
            exit_index = exit_index_map[exit_code]
            adj_count = int(count * multiplier)
            exit_count_list[exit_index] = adj_count
        # add the total to the count list
        try:
            cpu, sw = cpu_sw_map[task_name]
        except KeyError:
            # this task has no context switch no cpu, so likely is
            # using up all the cpu
            print 'keyerror:' + task_name
            cpu = 100
            sw = 1
        task_list.append({'name': task_name,
                          'exit_count': str(exit_count_list),
                          'cpu': round(cpu, 1),
                          'sw': sw})

    # get in reverse order so we display them top to bottom on a
    # horizontal stacked bar chart
    task_list.reverse()
    return (task_list, exit_reason_list, colormap_list)



#!/usr/bin/env python
# Copyright 2015 Cisco Systems, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
#
# Author: Alec Hothan
# ---------------------------------------------------------


import itertools
from pandas import DataFrame
import pandas
import numpy as np

from bokeh.plotting import figure
from bokeh.models.sources import ColumnDataSource
from bokeh.palettes import Spectral6

from perfmap_common import title_style
from perfmap_common import output_html
from perfmap_common import get_disc_size
from perfmap_common import get_time_span_usec

# For sorting
# 'CSR.1.vcpu0' => '0001.CSR.vcpu0'
def normalize_task_name(task):
    tokens = task.split('.')
    if len(tokens) > 1:
        # extract the service chain
        chain = tokens[1]
        try:
            chain = int(tokens[1])
            res = '%02d' % (chain)
            tokens.pop(1)
            return res + '.' + '.'.join(tokens)
        except ValueError:
            pass
    return task

def normalize_df_task_name(df):
    df['task_name'] = df.apply(lambda row: normalize_task_name(row['task_name']), axis=1)

def cycle_colors(chunk, palette=Spectral6):
    """ Build a color list just cycling through a given palette.

    Args:
        chunck (seq): the chunk of elements to generate the color list
        palette (seq[color]) : a palette of colors to cycle through

    Returns:
        colors

    """
    colors = []

    g = itertools.cycle(palette)
    for i in range(len(chunk)):
        colors.append(next(g))

    return colors

def show_core_locality(df, task_re, label):

    df = df[df['event'] == 'sched__sched_switch']
    df = df[df['task_name'].str.match(task_re)]
    # aggregate all the per core tasks (e.g. swapper/0 -> swapper)
    df['task_name'] = df['task_name'].str.replace(r'/.*$', '')
    # group by task name
    gb = df.groupby('task_name')
    task_list = gb.groups

    p = figure(plot_width=1000, plot_height=800, **title_style)

    p.xaxis.axis_label = 'time (usecs)'
    p.yaxis.axis_label = 'core'
    p.legend.orientation = "bottom_right"
    p.xaxis.axis_label_text_font_size = "10pt"
    p.yaxis.axis_label_text_font_size = "10pt"
    p.title = "Core locality (%s)" % (label)

    color_list = cycle_colors(task_list)

    for task, color in zip(task_list, color_list):
        dfe = gb.get_group(task)
        # add 1 column to contain the starting time for each run period
        dfe['start'] = dfe['usecs'] - dfe['duration']
        tid = dfe['pid'].iloc[0]
        count = len(dfe)
        legend_text = '%s:%d (%d)' % (task, tid, count)
        # draw end of runs
        p.circle('usecs', 'cpu', source=ColumnDataSource(dfe),
                 size=get_disc_size(count) + 2, color=color,
                 alpha=0.3,
                 legend=legend_text)
        # draw segments to show the entire runs
        p.segment('start', 'cpu', 'usecs', 'cpu', line_width=5, line_color=color,
                  source=ColumnDataSource(dfe))

    # specify how to output the plot(s)
    output_html(p, 'coreloc', task_re)

def filter_df_core(df, task_re, remove_cpu=False):
    # remove unneeded columns
    df = df.drop(['next_pid', 'pid', 'usecs', 'next_comm'], axis=1)
    if remove_cpu:
        df = df.drop('cpu', axis=1)

    # filter out all events except the switch events
    df = df[df.event == 'sched__sched_switch']
    df = df.drop('event', axis=1)
    df = df[df['task_name'].str.match(task_re)]
    return df

def get_coremaps(dfds, cap_time_usec, task_re):
    '''
    coremaps =  [
        { "run":"run1", "coremap": [
             { "task":"vnf1", "counts":[[0,5.5,231], [1,15.5,231], [2,25.5,21],
                                        [3,35.5,231],[4,45.8,415], [5,55.5,231],
                                        [6,65,231], [7,75.5,231],[8,85.5,231],
                                        [9,95.5,231]]},
             { "task":"vnf2", "counts":[[2,70.5,91], [31,45.1,152]]},
             { "task":"all tasks", "counts":[[1,10.5,231], [2,30.5,21], [4,21.8,415],[31,45.1,152]]}
             ],
             "minsw": 5, "maxsw": 415
        },
        { "run":"run2", "coremap": [
             { "task":"vnf1", "counts":[[4,40.5,51], [7,0.8,215]]},
             { "task":"vnf2", "counts":[[30,35.1,62]]},
             { "task":"all tasks", "counts":[[4,40.5,51], [7,0.8,215],[30,35.1,62]]
             ],
             "minsw": 51, "maxsw": 215
        }
    ]
    '''
    coremaps = []
    for dfd in dfds:
        df = dfd.df
        time_span_usec = get_time_span_usec(df)

        # remove unneeded columns
        df = filter_df_core(df, task_re)

        # at this point we have a df that looks like this:
        #         task_name  cpu  duration
        # 0     ASA.1.vcpu0    8      7954
        # 1     ASA.1.vcpu0    9      5475
        # 2     ASA.1.vcpu0    9      1000
        # 3     ASA.1.vcpu0   11     12391
        # 4     ASA.1.vcpu0   12     21025
        # 5     ASA.1.vcpu0   13      6447
        # etc...
        if len(df) == 0:
            print
            print 'No selection matching "%s"' % (task_re)
            return None
        gb = df.groupby(['task_name', 'cpu'], as_index=False)

        # because we only show percentages, there is no need to apply the multiplier
        # add duration values
        df = gb.aggregate(np.sum)
        #         task_name  cpu  duration
        # 0     ASA.1.vcpu0    8      7954
        # 1     ASA.1.vcpu0    9      6475
        # 2     ASA.1.vcpu0   11     12391

        # Add a % column
        df['percent'] = np.round((df['duration'] * 100) / time_span_usec, 2)

        # many core-pinned system tasks have a duration of 0 (swapper, watchdog...)
        df.fillna(100, inplace=True)
        df.drop(['duration'], axis=1, inplace=True)

        # count number of rows with same task and cpu
        dfsw = DataFrame(gb.size())
        dfsw.reset_index(inplace=True)
        dfsw.rename(columns={0: 'count'}, inplace=True)
        # adjust context switch count if the requested cap time is > time_span
        if dfd.multiplier > 1.0:
            dfsw['count'] = (dfsw['count'].astype(int) * dfd.multiplier).astype(int)
        min_count = dfsw['count'].min()
        max_count = dfsw['count'].max()
        #       task_name  cpu  count
        # 0  ASA.01.vcpu0    8   9853
        # 1  ASA.01.vcpu0    9    348
        # 1  ASA.01.vcpu0   11    619

        # Merge the 2 df using the task_name/cpu as the joining key
        dfm = pandas.merge(df, dfsw, how="left", on=['task_name', 'cpu'])

        alltasks_label = 'all tasks'

        # calculate the sum of all percent and switches per task
        dfallcores = dfm.drop('cpu', axis=1)
        gball = dfallcores.groupby('task_name')
        dfallcores = gball.aggregate(np.sum)
        dfallcores['percent'] = np.round(dfallcores['percent'], 2)
        dfallcores['task_name'] = dfallcores.index
        dfallcores['cpu'] = '0-31'

        # sort all the tasks in reverse order (without the 'all tasks')
        task_list = gball.groups.keys()
        task_list.sort(reverse=True)
        task_list.append(alltasks_label)
        # concatenate the 2 dfs
        dfm = pandas.concat([dfm, dfallcores], ignore_index=True)

        # calculate the sum of all percent and switches per cpu
        dfalltasks = dfm.drop('task_name', axis=1)
        gball = dfalltasks.groupby('cpu')
        dfalltasks = gball.aggregate(np.sum)
        dfalltasks['task_name'] = alltasks_label
        dfalltasks['cpu'] = dfalltasks.index
        dfalltasks['percent'] = np.round(dfalltasks['percent'], 2)
        # concatenate all dfs
        dfm = pandas.concat([dfm, dfalltasks], ignore_index=True)

        # generate the data structure for the jinja template
        cml = []
        gb = dfm.groupby('task_name')

        for task in task_list:
            counts = []
            dfg = gb.get_group(task)
            for index, row in dfg.iterrows():
                counts.append([row['cpu'], row['percent'], row['count']])
            cml.append({"task": task, "counts": counts})

        coremap = {"run": dfd.short_name, "coremap": cml, "extent": str([min_count, max_count])}
        coremaps.append(coremap)
    return coremaps

def get_cpu_sw_map(dfds, cap_time_usec, task_re):
    df_list = []
    dfsw_list = []
    for dfd in dfds:
        df = filter_df_core(dfd.df, task_re, True)
        # at this point we have a set of df that look like this:
        #         task_name  duration
        # 0     ASA.1.vcpu0      7954
        # 1     ASA.1.vcpu0      5475
        # 2     ASA.1.vcpu0      4151
        if df.empty:
            continue
        gb = df.groupby('task_name', as_index=False)

        # sum all duration for each task
        df = gb.aggregate(np.sum)
        if dfd.multiplier > 1.0:
            df['duration'] = (df['duration'] * dfd.multiplier).astype(int)
        df['percent'] = ((df['duration'] * 100 * 10) // cap_time_usec) / 10
        if len(dfds) > 1:
            df['task_name'] = df['task_name'] + '.' + dfd.short_name
        df_list.append(df)

        # count number of rows with same task and cpu
        dfsw = DataFrame(gb.size())
        dfsw.reset_index(inplace=True)
        dfsw.rename(columns={0: 'count'}, inplace=True)

        if dfd.multiplier > 1.0:
            dfsw['count'] = (dfsw['count'] * dfd.multiplier).astype(int)
        dfsw_list.append(dfsw)

    df = pandas.concat(df_list)
    df = df.drop('duration', axis=1)
    dfsw = pandas.concat(dfsw_list)
    df = pandas.merge(df, dfsw, on='task_name')
    # Result:
    #             task_name  percent  count
    # 0  ASA.01.vcpu0.1x218     72.0  1998
    # 1  ASA.01.vcpu0.2x208     61.8  2128
    # 2  ASA.02.vcpu0.2x208     58.9  2177

    # transform this into a dict where the key is the task_name and the value
    # is a list [percent, count]
    return df.set_index('task_name').T.to_dict('list')



#!/usr/bin/env python
# Copyright 2015 Cisco Systems, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
#
# Author: Alec Hothan
# ---------------------------------------------------------

# This is an example of plugin that will load the list of instances from OpenStack
# and extract the type of service and service chain name from the instance name
# and store the resulting descriptive string in a dictionary indexed by the uuid
#

import credentials
import os
import re
from novaclient.client import Client

# Config file to specify the OpenStack crendentials needed to connect to the controller
# The file must contain the rc variable to point to the OpenStack credentials file
# (as downloaded from the Horizon dashboard), for example:
# rc=../admin-oper.sh
CFG_FILE = '.mkcdict.cfg'

# Extract the VM type and service chain ID from the instance name
# ESC_Day0-3__68540__MT__MTPerftest-FULL-01ESC_Day0-31.1__0__ASA__0
# ESC_Day0-3__62940__MT__MTPerftest_FULL_01ESC_Day0-31.1__0__CSR__0
instance_re = re.compile('ESC_Day0-3__\d*__MT__MTPerftest.FULL.(\d*)ESC_Day0-31.1__0__([A-Z]*)_')

# A dict of full names indexed by the uuid
#
# ESC_Day0-3__62940__MT__MTPerftest_FULL_01ESC_Day0-31.1__0__CSR__0
# becomes 'CSR.01'
by_uuid = {}

def decode_instance_name(name):
    m = instance_re.match(name)
    if m:
        chain_id = m.group(1)
        nvf = m.group(2)
        return int(chain_id), nvf
    return None, None


class OptionsHolder(object):
    def __init__(self):
        # check if there is a config file in the current directory
        if os.path.isfile(CFG_FILE):
            # load options from the config file
            opt_re = re.compile(' *(\w*) *[=:]+ *([\w/\-\.]*)')
            with open(CFG_FILE, 'r') as ff:
                for line in ff:
                    m = opt_re.match(line)
                    if m:
                        setattr(self, m.group(1), m.group(2))


def plugin_init(opts=None):
    # Parse the credentials of the OpenStack cloud
    if not opts:
        opts = OptionsHolder()
    cred = credentials.Credentials(opts)
    creds_nova = cred.get_nova_credentials_v2()
    # Create the nova and neutron instances
    nova_client = Client(**creds_nova)

    opts = {'all_tenants': 1}
    servers = nova_client.servers.list(detailed=True, search_opts=opts)
    count = 0
    for server in servers:
        chain_id, nvf = decode_instance_name(server.name)
        if chain_id:
            full_name = '%s.%02d' % (nvf, chain_id)
            by_uuid[server.id] = full_name
            count += 1
    print 'Plugin loaded with %d service chain names from Nova' % (count)
    return True

def plugin_convert_name(name, tid, libvirt_name, uuid, thread_type):
    try:
        return by_uuid[uuid]
    except KeyError:
        by_uuid[uuid] = name
        return name



#!/usr/bin/env python
# Copyright 2015 Cisco Systems, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
#
# Author: Alec Hothan
# ---------------------------------------------------------


from optparse import OptionParser
import os
import sys
import warnings
import pandas
from pandas import DataFrame

from perf_formatter import open_cdict

from perfmap_common import set_html_file
from perfmap_common import DfDesc
from perfmap_common import output_svg_html
from perfmap_core import get_coremaps
from perfmap_core import show_core_locality
from perfmap_kvm_exit_types import get_kvm_exit_data
from perfmap_sw_kvm_exits import show_sw_kvm_heatmap

from jinja2 import Environment
from jinja2 import FileSystemLoader
import time
from __init__ import __version__
from pkg_resources import resource_filename

# Global variables
output_chart = None

# start analysis after first from_time usec
from_time = 0
# cap input file to first cap_time usec, 0 = unlimited
cap_time = 0

def get_full_task_name(df, task):
    # if task is a number it is considered to be a pid ID
    # if text it is a task name
    try:
        tid = int(task)
        # tid given
        df = df[df['pid'] == tid]
        task = None
        # get the task name from the tid
    except ValueError:
        # task given: find corresponding tid
        df = df[df['task_name'] == task]
        tid = 0

    if df.empty:
        print 'No selection matching the task ' + task
        return (None, None)
    # fill in the missing information
    if not tid:
        tid = df['pid'].iloc[0]
    if not task:
        task = df['task_name'].iloc[0]
    task = task + ':' + str(tid)
    return (df, task)

def show_successors(df, task, label):
    df, task = get_full_task_name(df, task)
    if not task:
        return
    df = df[df['event'] == 'sched__sched_switch']
    # aggregate all the per core tasks (e.g. swapper/0 -> swapper)
    df['next_comm'] = df['next_comm'].str.replace(r'/.*$', '')
    series_percent = df.next_comm.value_counts(normalize=True)
    series_count = df.next_comm.value_counts()
    series_percent = pandas.Series(["{0:.2f}%".format(val * 100) for val in series_percent],
                                   index=series_percent.index)

    series_percent.name = 'percent'
    series_count.name = 'count'
    print 'Successors of %s (%s)' % (task, label)
    print pandas.concat([series_count, series_percent], axis=1)

def set_short_names(dfds):
    '''
    Reduce the names of a list of dataframe descriptors to minimal non matching characters
    This will basically trim from the start and end all common strings and keep only the
    non matching part of the names.
    Example of names: ../../haswell/h1x216.cdict    ../../haswell/h5x113.cdict
    Resulting reduced names: h1x216 h5x113
    Caveat: do not strip out any numeric digits at both ends, e.g. 01x218.cdict 02x208.cdict
    must return [1x218, 2x208] must not cut out the trailing 8

    '''
    name_list = [dfd.name for dfd in dfds]
    if len(name_list) < 2:
        return
    strip_head = None
    strip_tail = None
    for name in name_list:
        if not strip_head:
            strip_head = name
            strip_tail = name
            continue
        # because there are no duplicates we know that
        # at least 1 character difference between all keys
        # find longest match from head
        max_index = min(len(name), len(strip_head))
        for index in range(max_index):
            if name[index] != strip_head[index] or name[index].isdigit():
                strip_head = name[:index]
                break

        # find longest match from tail
        max_index = min(len(name), len(strip_tail))
        for index in range(-1, -max_index - 1, -1):
            # stop when mismatch or when hitting a numeric letter
            if name[index] != strip_tail[index] or name[index].isdigit():
                if index == -1:
                    strip_tail = ''
                else:
                    strip_tail = name[1 + index:]
                break

    # strip all names and store in the short_name field
    for dfd in dfds:
        dfd.short_name = dfd.name[len(strip_head):]
        if strip_tail:
            dfd.short_name = dfd.short_name[:-len(strip_tail)]


def create_charts(dfds, cap_time_usec, task_re, label):

    coremaps = get_coremaps(dfds, cap_time_usec, task_re)

    task_list, exit_reason_list, colormap_list = get_kvm_exit_data(dfds, cap_time_usec, task_re)

    # Other misc information in the chart
    info = {
        "label": label,
        "window": "{:,d}".format(cap_time_usec / 1000),
        "date": time.strftime("%d-%b-%Y"),    # 01-Jan-2016 format
        "max_cores": 32,
        "version": __version__
    }
    local_path = resource_filename(__name__, 'perfmap_charts.jinja')
    print local_path
    template_loader = FileSystemLoader(searchpath=".")
    template_env = Environment(loader=template_loader, trim_blocks=True, lstrip_blocks=True)
    tpl = template_env.get_template(local_path)
    svg_html = tpl.render(exit_reason_list=str(exit_reason_list),
                          task_list=task_list,
                          colormap_list=str(colormap_list),
                          coremaps=coremaps,
                          info=info)
    output_svg_html(svg_html, 'charts', task_re)

# ---------------------------------- MAIN -----------------------------------------

def main():
    global from_time
    global cap_time

    # Suppress future warnings
    warnings.simplefilter(action='ignore', category=FutureWarning)

    parser = OptionParser(usage="usage: %prog [options] <cdict_file1> [cdict_file2...]")

    parser.add_option("--successors-of",
                      dest="successor_of_task",
                      help="show list of successors of given tid or task name"
                      )
    parser.add_option("--show-tids",
                      dest="show_tids",
                      action="store_true",
                      default=False,
                      help="show list of all tids with event count)"
                      )
    parser.add_option("-t", "--task",
                      dest="task",
                      metavar="task name (regex)",
                      help="selected task(s) (regex on task name)"
                      )
    parser.add_option("--core-locality",
                      dest="core_locality",
                      action="store_true",
                      help="show core locality heat map (requires --task)"
                      )
    parser.add_option("--switches",
                      dest="switches",
                      action="store_true",
                      help="show context switches heat map (requires --task)"
                      )
    parser.add_option("--kvm-exits",
                      dest="kvm_exits",
                      action="store_true",
                      help="show kvm exits heat map (requires --task)"
                      )
    parser.add_option("--show-sleeps",
                      dest="show_sleeps",
                      action="store_true",
                      help="also show sleep events in the context switch heat map (defaults: hide)"
                      )
    parser.add_option("--label",
                      dest="label",
                      metavar="label",
                      help="label for the title (defaults to the cdict file name)"
                      )
    parser.add_option("--map",
                      dest="map",
                      action="store",
                      metavar="mapping csv file",
                      help="remap task names from mapping csv file"
                      )
    parser.add_option("--output-dir",
                      dest="output_dir",
                      action="store",
                      metavar="dir pathname",
                      help="output all html files to the provided directory"
                      )
    parser.add_option("--headless",
                      dest="headless",
                      action="store_true",
                      help="do not show chart in the browser (default=False)"
                      )
    parser.add_option("-c", "--cap",
                      dest="cap_time",
                      help="(optional) cap the analysis to first <cap_time> msec"
                           " of capture (default=all)"
                      )
    parser.add_option("-f", "--from",
                      dest="from_time",
                      help="(optional) start the analysis after first <from_time> msec"
                           " of capture (default=0)"
                      )
    (options, args) = parser.parse_args()

    if options.from_time:
        from_time = int(options.from_time) * 1000
    if options.cap_time:
        # convert to usec
        cap_time = int(options.cap_time) * 1000 + from_time

    if not args:
        print 'Missing cdict file'
        sys.exit(1)

    if options.output_dir:
        if not os.path.isdir(options.output_dir):
            print('Invalid output directory: ' + options.output_dir)
            sys.exit(1)

    dfds = []
    cdict_files = args
    if len(cdict_files):
        if len(cdict_files) > 1:
            html_filename = cdict_files[0] + '-diff'
        else:
            html_filename = cdict_files[0]
    else:
        html_filename = cdict_files[0] if len(cdict_files) else 'perfwhiz'
    set_html_file(html_filename, options.headless, options.label, options.output_dir)

    # get smallest capture window of all cdicts
    min_cap_usec = 0
    for cdict_file in cdict_files:
        perf_dict = open_cdict(cdict_file, options.map)
        df = DataFrame(perf_dict)
        dfd = DfDesc(cdict_file, df)
        dfds.append(dfd)
        last_usec = df['usecs'].iloc[-1]
        if min_cap_usec == 0:
            min_cap_usec = last_usec
        else:
            min_cap_usec = min(min_cap_usec, last_usec)
    if from_time and from_time >= min_cap_usec:
        print 'Error: from time cannot be larger than %d msec' % (min_cap_usec / 1000)
        sys.exit(2)
    # if a cap time is provided, always use that value (even if it is >min_cap_usec)
    if not cap_time:
        cap_time = min_cap_usec

    # normalize all dataframes
    for dfd in dfds:
        dfd.normalize(from_time, cap_time)

    # at this point some cdict entries may have "missing" data
    # if the requested cap_time is > the cdict cap time
    # the relevant processing will extrapolate when needed (and if possible)

    # reduce all names to minimize the length of the cdict file name
    set_short_names(dfds)

    if not options.label:
        if len(dfds) > 1:
            options.label = 'diff'
        else:
            options.label = os.path.splitext(os.path.basename(cdict_file))[0]

    if options.show_tids:
        print 'List of tids and task names sorted by context switches and kvm event count'
        for dfd in dfds:
            print dfd.name + ':'
            res = dfd.df.groupby(['pid', 'task_name']).size()
            res.sort_values(ascending=False, inplace=True)
            print res
        sys.exit(0)

    if options.successor_of_task:
        for dfd in dfds:
            print dfd.name + ':'
            show_successors(dfd.df, options.successor_of_task, options.label)
        sys.exit(0)

    # These options can be cumulative and all require a --task parameter to select tasks
    if not options.task:
        print '--task <task_regex> is required'
        sys.exit(1)

    if options.core_locality:
        if len(dfds) == 1:
            show_core_locality(dfds[0].df, options.task, options.label)
            sys.exit(0)
        else:
            print 'Core locality diff is not supported - can only accept 1 cdict argument'
            sys.exit(1)

    if options.switches or options.kvm_exits:
        show_sw_kvm_heatmap(dfds[0].df, options.task, options.label, options.switches, options.kvm_exits,
                            options.show_sleeps)

    create_charts(dfds, cap_time, options.task, options.label)

if __name__ == '__main__':
    main()



# Copyright 2015 Cisco Systems, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

# Common functions across capture and map functions

import csv
import marshal
import os
import re
import zlib
try:
    # try to use the faster version if available
    from msgpack import packb
    from msgpack import unpackb
except ImportError:
    # else fall back to the pure python version (slower)
    from umsgpack import packb
    from umsgpack import unpackb

# a dict of task names indexed by tid
name_by_tid = {}
plugin_convert_name = None

def init(opts):
    global plugin_convert_name
    # try to import, can raise ImportError (no plugin found)
    # or plugin_init can raise ValueError
    from mkcdict_plugin import plugin_init
    print 'Initializing plugin...'
    if plugin_init(opts):
        from mkcdict_plugin import plugin_convert_name
        print 'Plugin initialized successfully'

uuid_re = re.compile('-uuid ([a-fA-F0-9\-]*)')
# /proc/pid/cpuset output
# /machine/instance-000065d7.libvirt-qemu/emulator
# /machine/instance-000065d3.libvirt-qemu/vcpu0
cpuset_re = re.compile('/machine/(instance-[a-fA-F0-9]*).libvirt-qemu/(\w*)')

def decode_cpuset(pid):
    name = None
    thread_type = None
    try:
        with open('/proc/%d/cpuset' % (pid)) as f:
            cpuset = f.read()
            res = cpuset_re.match(cpuset)
            if res:
                name = res.group(1)
                thread_type = res.group(2)
    except IOError:
        # some pids/tids come and go so just forget about this one
        pass
    return name, thread_type

def decode_pid(pid):
    name = None
    uuid = None
    thread_type = None
    try:
        with open('/proc/%d/cmdline' % (pid)) as f:
            cmdline = f.read()
            # this is a nul separated tokens
            cmdline = cmdline.replace('\x00', ' ')
            if cmdline.startswith('/usr/bin/qemu-system'):
                res = uuid_re.search(cmdline)
                if res:
                    uuid = res.group(1)
                    name, thread_type = decode_cpuset(pid)
    except IOError:
        # some pids/tids come and go so just forget about this one
        pass
    return name, uuid, thread_type

def get_task_name(tid, name):
    if not tid:
        return name
    try:
        return name_by_tid[tid]
    except KeyError:
        pass
    # check if it is a kvm thread from the look of the name
    libvirt_name, uuid, thread_type = decode_pid(tid)
    if libvirt_name:
        if plugin_convert_name:
            name = plugin_convert_name(name, tid, libvirt_name, uuid, thread_type)
        # append the thread type to the name
        name += '.' + thread_type
    name_by_tid[tid] = name
    return name

# cdict management functions

def remap(perf_dict, csv_map):
    '''Remap all the task names in the cdict file with those specified in the mapping file
    :param perf_dict: an uncompressed dictionary
    :param csv_map: csv mapping file name
    '''
    # a mapping dict of task names indexed by the tid
    map_dict = {}
    print 'Remapping task names...'
    with open(csv_map, 'r') as ff:
        # 19236,instance-000019f4,emulator,8f81e3a1-3ebd-4015-bbee-e291f0672d02,FULL,5,CSR
        reader = csv.DictReader(ff, fieldnames=['tid', 'libvirt_id', 'thread_type', 'uuid', 'chain_type',
                                                'chain_id', 'nvf_name'])
        for row in reader:
            task_name = '%s.%02d.%s' % (row['nvf_name'], int(row['chain_id']), row['thread_type'])
            map_dict[int(row['tid'])] = task_name
    pids = perf_dict['pid']
    names = perf_dict['task_name']
    next_pids = perf_dict['next_pid']
    next_comms = perf_dict['next_comm']
    count = 0
    for index in xrange(len(pids)):
        try:
            new_task_name = map_dict[pids[index]]
            names[index] = new_task_name
            count += 1
        except KeyError:
            pass
        try:
            new_task_name = map_dict[next_pids[index]]
            next_comms[index] = new_task_name
            count += 1
        except KeyError:
            pass
    print 'Remapped %d task names' % (count)

def open_cdict(cdict_file, map_file=None):
    '''Open and decode a cdict file
    :param cdict_file: name of the cdict file
    :param map_file: name of a mapping file (optional)
    :return: the uncompressed dictionary representing the cdict file
    '''
    if not cdict_file.endswith('.cdict'):
        # automatically add the cdict extension if there is one
        if os.path.isfile(cdict_file + '.cdict'):
            cdict_file += '.cdict'
        else:
            raise ValueError('cdict file name must have the .cdict extension: ' + cdict_file)

    with open(cdict_file, 'r') as ff:
        cdict = ff.read()

    decomp = zlib.decompress(cdict)
    try:
        perf_dict = unpackb(decomp)
    except Exception:
        # old serialization format
        perf_dict = marshal.loads(decomp)
    if map_file:
        remap(perf_dict, map_file)
    return perf_dict

def write_cdict(cdict_file, perf_dict):
    '''Write a dictionary to a cdict file
    :param cdict_file: cdict file name (will auto add a .cdict extension if missing)
    :param perf_dict:  perf dict to compress and write
    :return:
    '''
    if not cdict_file.endswith('.cdict'):
        # automatically add the cdict extension if there is one
        cdict_file += '.cdict'
    with open(cdict_file, 'w') as ff:
        compressed = zlib.compress(packb(perf_dict))
        ff.write(compressed)
        print 'Compressed dictionary written to %s %d entries size=%d bytes' % \
              (cdict_file, len(perf_dict), len(compressed))



#!/usr/bin/env python
# Copyright 2015 Cisco Systems, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
#
# Author: Alec Hothan
# ---------------------------------------------------------

import sys

from bokeh.plotting import figure
from bokeh.models.sources import ColumnDataSource
from bokeh.models import Range1d
from bokeh.io import gridplot

from perfmap_common import title_style
from perfmap_common import grid_title_style
from perfmap_common import output_html
from perfmap_common import get_disc_size

GREEN = "#5ab738"
RED = "#f22c40"
BLUE = "#4169E1"
ORANGE = "#FFA500"

def get_groupby(df, task_re):
    # if task is a number it is considered to be a pid ID
    # if text it is a task name
    try:
        tid = int(task_re)
        # tid given
        df = df[df['pid'] == tid]
        # a groupby with only 1 group since we already filtered on the pid
        gb = df.groupby('pid')
    except ValueError:
        # task given: find corresponding tid
        df = df[df['task_name'].str.match(task_re)]
        gb = df.groupby('task_name')

    return gb

def split_list(l, n):
    # split a list into multiple lists of given len
    n = max(1, n)
    return [l[i:i + n] for i in range(0, len(l), n)]

def show_sw_kvm_heatmap(df, task_re, label, show_ctx_switches, show_kvm, show_sleeps):
    gb = get_groupby(df, task_re)

    chart_list = []
    # these are the 2 main events to show for kvm events
    legend_map_kvm = {
        'kvm_exit': (BLUE, 'vcpu running (y=vcpu run time)', False),
        'kvm_entry': (ORANGE, 'vcpu not running (y=kvm+sleep time)', False)
    }
    # context switch events
    legend_map_ctx_sw = {
        'sched__sched_switch': (GREEN, 'switched out from cpu (y=run time)', True)
    }
    if show_sleeps:
        # note that some versions of Linux do not generate sched__sched_stat_sleep traces
        legend_map_ctx_sw['sched__sched_stat_sleep'] = (RED, 'wakeup from sleep (y=sleep time)', True)
    if show_kvm and show_ctx_switches:
        legend_map = dict(legend_map_kvm.items() + legend_map_ctx_sw.items())
        title = "Scheduler and KVM events"
        prefix = 'swkvm'
    elif show_kvm:
        legend_map = legend_map_kvm
        title = "KVM events"
        prefix = 'kvm'
    else:
        legend_map = legend_map_ctx_sw
        title = "Scheduler events"
        prefix = 'sw'
    width = 1000
    height = 800
    show_legend = True
    nb_charts = len(gb.groups)
    if nb_charts == 0:
        print 'No selection matching: ' + task_re
        return
    if nb_charts > 1:
        width /= 2
        height /= 2
        tstyle = grid_title_style
    else:
        tstyle = title_style
    task_list = gb.groups.keys()
    task_list.sort()
    show_legend = True
    duration_max = usecs_max = -1
    duration_min = usecs_min = sys.maxint
    event_list = legend_map.keys()
    event_list.sort()

    for task in task_list:
        p = figure(plot_width=width, plot_height=height, y_axis_type="log", **tstyle)
        p.xaxis.axis_label = 'time (usecs)'
        p.yaxis.axis_label = 'duration (usecs)'
        p.legend.orientation = "bottom_right"
        p.xaxis.axis_label_text_font_size = "10pt"
        p.yaxis.axis_label_text_font_size = "10pt"
        if label:
            p.title = "%s for %s (%s)" % (title, task, label)
            label = None
        else:
            p.title = task
        p.ygrid.minor_grid_line_color = 'navy'
        p.ygrid.minor_grid_line_alpha = 0.1
        accumulated_time = {}
        total_time = 0

        dfg = gb.get_group(task)
        # remove any row with zero duration as it confuses the chart library
        dfg = dfg[dfg['duration'] > 0]

        for event in event_list:
            dfe = dfg[dfg.event == event]
            duration_min = min(duration_min, dfe['duration'].min())
            duration_max = max(duration_max, dfe['duration'].max())
            usecs_min = min(usecs_min, dfe['usecs'].min())
            usecs_max = max(usecs_max, dfe['usecs'].max())
            count = len(dfe)
            color, legend_text, cx_sw = legend_map[event]
            if show_legend:
                legend_text = '%s (%d)' % (legend_text, count)
            elif color == GREEN:
                legend_text = '(%d)' % (count)
            else:
                legend_text = None
            # there is bug in bokeh when there are too many circles to draw, nothing is visible
            if len(dfe) > 50000:
                dfe = dfe[:50000]
                print 'Series for %s display truncated to 50000 events' % (event)
            if cx_sw:
                draw_shape = p.circle
                size = get_disc_size(count)
            else:
                draw_shape = p.diamond
                size = get_disc_size(count) + 4

            draw_shape('usecs', 'duration', source=ColumnDataSource(dfe),
                       size=size,
                       color=color,
                       alpha=0.3,
                       legend=legend_text)
            event_duration = dfe['duration'].sum()
            accumulated_time[event] = event_duration
            total_time += event_duration
        chart_list.append(p)
        show_legend = False

    shared_x_range = Range1d(usecs_min, usecs_max)
    shared_y_range = Range1d(duration_min, duration_max)

    for p in chart_list:
        p.x_range = shared_x_range
        p.y_range = shared_y_range

    # specify how to output the plot(s)

    # display the figure
    if len(chart_list) == 1:
        output_html(chart_list[0], prefix, task_re)
    else:
        # split the list into an array of rows with 2 charts per row
        gp = gridplot(split_list(chart_list, 2))
        output_html(gp, prefix, task_re)
